#include <port.h>

/*****************************************************************************/
/*      ucmain.c		          				     */
/*---------------------------------------------------------------------------*/
/*	Main section of code for the canny edge detector		     */
/*---------------------------------------------------------------------------*/
/*
    PURPOSE:
       Detect edge points in an image.
       
    INPUT: (Non-interactive)
       - Pointer to the image data in row-major order
         (256 grey-levels, 1 byte per pixel)
       - Width of image
       - Height of image
       - Standard Deviation (sigma) of gaussian filter
       - Low  gradient threshold for edge labeling
       - high gradient threshold for edge labeling
       - Pointer to the pointer to the edge list structure to return
       
    COMMENTS:
									     */
/*****************************************************************************/

/* INCLUDE FILES */
#include "uc.h"
#include "sys/types.h"
#include "imgformat.h"


/* Global variables */
PGMImage	*pgmimage ;

double		sigma          = 0.4 ;
int		mask           = 3 ;
float low_threshold  = 1.4 ;
float high_threshold = 2.0 ;
int		XSIZE          = 128 ;
int		YSIZE          = 128 ;
int		Sub_pixel      = FALSE ;
char		outfile[MAXPATHLEN] ;	/* defalut output file name */
char		infile[MAXPATHLEN] ;	/* defalut output file name */

/* Statistical counters (more Globals) */
int		NUM_JCTS   = 0 ;
int		NUM_EDGELS = 0 ;
int		NUM_EDGES  = 0 ;

unsigned short	*Record_image = NULL ;
unsigned short	*image        = NULL ;
unsigned short	*timage       = NULL ;

/* Hash table of edgels */
edgel **edgels = NULL ;
edgel *tedgels = NULL ;
edge *edges = NULL ;
edge *tedges = NULL ;

args_t args[] = {
  ARG_DEF("i", TRUE, STRING, MAXPATHLEN, infile,
          "the name of the PGM image file for input"),
  ARG_DEF("s", FALSE, FLOAT, 1, &sigma,
          "the sigma value used for convolution (default 0.4)"),
  ARG_DEF("m", FALSE, INT, 1, &mask,
          "the size of the full mask used for convolution (default 4*sigma)"),
  ARG_DEF("lt", FALSE, FLOAT, 1, &low_threshold,
          "the low threshold value (default 1.4)"),
  ARG_DEF("ht", FALSE, FLOAT, 1, &high_threshold,
          "the low threshold value (default 2.0)"),
  ARG_DEF("sp", FALSE, BOOLEAN, 1, &Sub_pixel,
          "print out sub-pixel localized values (default FALSE)"),
  ARG_DEF("o", TRUE, STRING, MAXPATHLEN, outfile,
          "the name of the file to store the connectivity list"),
  ARG_LAST
  } ;                                                                                                                                                                                                             


void canny() ;

/* External functions */
extern void convolve() ;
extern void local_extrema() ;
extern void join_edges() ;
extern void write_file() ;
extern void args_parse() ;

/* External utility functions */
extern PGMImage *read_pgm_file() ;
extern void write_pgm_file() ;
extern void free_pgm_data() ;


void main(int argc, char **argv)
  {
  /* Parse the command-line arguments */
  args_parse(argc, argv, args) ;

  /* Make sure the mask size is the smallest odd number greater
     than six times the s.d. */
  mask = MAX( (int)(MASK_SIGMAS * sigma + 0.5), 3 ) ;
  if (( mask % 2) == 0)
    mask++ ;
	
  /* Load the image appropriately */
  pgmimage = read_pgm_file(infile) ;
  XSIZE = pgmimage->ImageWidth ;
  YSIZE = pgmimage->ImageHeight ;
	
  /* Initialize any global variables */
  edges = NULL ;

  canny() ;
  }


/* Routine to do a majority of the leg-work in implementing
   the edge detection */
void canny()
  {
  double	*gx, *gy ;	/* Gradient array for X and Y direction */
  double	*mag ;		/* Magnitude of gradient array */

  /* Convolve image with a 2-D discrete gaussian using an
     array of dimensions mask x mask, and a s.d. of sigma */
  /* Convolve returning with gx & gy as the X and Y gradient, and
     mag being the magnitude of that */
  convolve(pgmimage->data, mask, sigma, &gx, &gy, &mag) ;
  
  local_extrema(mask, gx, gy, mag, low_threshold, high_threshold) ;
  
  join_edges(gx, gy) ;

  /* Create the output */
  write_file(edges, outfile) ;
  
  /* Free up memory */
  free(gx) ;
  free(gy) ;
  free(mag) ;
  }